#pragma TextEncoding="UTF-8"
#pragma rtGlobals=3
#pragma version=4.30
#pragma IgorVersion=7
#pragma ModuleName=BaselineSpline

// Project Updater header
static constant kProjectID=345 // the project node on IgorExchange
static strconstant ksShortTitle="BaselineSpline" // the project short title on IgorExchange

// If you have installed or updated this project using the IgorExchange 
// Projects Installer (http: www.igorexchange.com/project/Updater) you 
// can be notified when new versions are released.

// Written by Tony Withers, https://www.wavemetrics.com/user/tony
// I would be very happy to hear from you if you find this package
// useful, or if you have any suggestions for improvement.

// How to use Baseline Spline Fit

// A 1d wave must be plotted in a graph window before selecting Spline 
// Baseline from the macros menu. A control panel will be added to the 
// left of the graph window. When 'edit mode' is selected you can drag 
// the nodes. The spline updates as the nodes are repositioned. When in 
// edit mode you can add a new node at any position on the graph by 
// positioning the mouse cursor and clicking while holding down the 
// control key. Nodes may be deleted by clicking them with the option key
// held down. Clicking anywhere in the graph window with the shift key 
// held down will toggle between node editing mode and 'normal' mode 
// where you can interact with the graph as usual. Cubic and smoothing 
// splines are calculated by Igor's Interpolate2 operation; execute 
// 'DisplayHelpTopic Interpolate2' for details of the methods of spline 
// interpolation.

// Panel controls

// Data wave: select a trace for baseline correction. When switching
// between traces the nodes will retain their X-positions but will be
// placed on the selected data wave at those positions (unless they lie
// outside of the range of the data wave).

// Akima: fit an Akima style spline through the nodes. Akima splines can
// have sharper bends than cubic splines.

// PCHIP: fit a cubic Hermite spline through the nodes. This spline type
// doesn't overshoot local maxima or minima in node sequence. First
// derivative is continuous through the nodes, but second derivative may
// be discontinuous.

// Cubic: fit a 'normal' cubic spline through the nodes. Second
// derivative is continuous though the nodes.

// Ends: At the ends of the spline where the cubic function is not fully
// constrained select a method to add a constraint. 1st deriv.: match 1st
// derivative; natural: match 2nd derivative. Affects only cubic spline
// type.

// Linear: connect the nodes with linear segments.

// Smoothing: fit a smoothing spline, which does not have to pass exactly
// through the nodes. The degree of smoothing is adjustable.

// Edit mode: toggle between node editing and normal modes

// Reset: set node positions by spacing a set number of nodes at regular
// intervals along the section of the data wave displayed in the graph
// window. No nodes will be set beyond the range of the graph axes.

// Load: reload the node positions from a previous fit.

// Subtract: creates a copy of the baseline and a baseline-subtracted
// output spectrum in the current data folder and appends them to the
// plot; saves the node positions in a 2d wave in the current data
// folder.


// Version history

// 4.30 3 June 2021 End of the line for this project. Development will be
// continued as part of the Baselines project:
// https://www.wavemetrics.com/project/Baselines
// 4.20 17 July 2020 Removed code used by older versions of Updater
// 4.15 1/3/20 waves can be plotted on any axes
// 4.14 corrects the EndDerivativePCHIP function for PCHIP splines.
// Previous versions may in some circumstances extrapolate beyond the end
// nodes in an unreasonable way.
// 4.11 2/26/19 Added headers for Project Updater, fixed wave indexing bug
// by switching to double precision precision node waves
// 4.10 2/2/18 Introduces piecewise cubic Hermite spline
// 4.03 11/11/17 Should be backward compatible for Igor 6.
// 4.02 Catch interpolate errors.
// 4.01 Panel size decreased for compatibility with smaller or lower
// resolution screens. Some bugs fixed.
// 4.00 Changed to panel interface. Major rewrite.
// 3.20 Removed option to prevent nodes from being inserted. Now the /NI
// no insert flag is used by default, and we handle node insertion by
// using a hook to collect ctrl-clicks.
// 3.10 Some changes to the global strings and variables in the package
// folder. Tried and failed to add improved guesses for node positions.
// 3.01 11/4/16 Bug fix - waves with deltaX<0 produced nodes that were
// reverse sorted in X, but Akima calculation cannot deal with reverse
// sorted nodes.
// 3.00 4/11/16 Added option for Akima style spline.
// 2.60 3/30/16 Added option to supress node addition. This is a toggle
// that switches on the /NI option in GraphWaveEdit. Updated a lot of
// other code. Changed some of the defaults (now saves nodes by default),
// so behaviour is not the same as previous version.
// 2.50 Log mode to save node positions to file. Can load node positions
// from a file.
// 2.40 Added textbox reminder of how to toggle graph mode, renamed
// procedures for consistency.
// 2.30 5/16/11 Bug fix - code was choking on non-existent wave in first
// call to clearGraph(). Added panel to change fitting options.
// 2.10 Call clearGraph() from subtractBaseline().
// 2.00 8/6/09 Incorporated suggestions from JJ Weimer that improve
// functionality and clarity of code and compatibility with 6.10 ... and
// then wrecked his nice programming with my clumsy code.
// 1.43b Allow refit function to switch interpolate flag when number of
// nodes is too small to allow linear fit.
// 1.42 6/8/08 Set y axis to manual range to avoid crazy rescaling if
// spline shoots out of range. Show subtracted spectrum whilst fitting.
// 1.41 10/9/07 For XY data we now interpolate directly into XY baseline.
// Improved guesses for initial node positions for unevenly spaced XY
// data.
// 1.40 10/9/07 Fixed bug that occurred with baseline subtraction of XY
// data.
// 1.30 7/24/07 Handles XY data. Baseline is waveform, but data wave need
// not be (though it should be reasonably closely spaced).
// 1.20 7/3/07 Uses new graphwaveedit flag. Now requires version 6.0.1.

static constant kRecolorTraces=1 // set to 0 to disable trace color changes
static constant kSetNodesAlgorithm=1 // 0 puts nodes on trace y value, 1 positions node Y value on a guess for baseline 
static constant kRemoveOriginal=0 // choose whether to remove data trace after baseline subtraction
static constant kAppendBaselines=0 // choose whether to append baseline trace after subtraction

static strconstant ksBaselineSuffix="_BL"
static strconstant ksSubtractedSuffix="_sub"

menu "Analysis"
	"Spline Baseline",/Q, BaselineSpline#Initialise()
end

//menu "Macros"
//	"Spline Baseline",/Q, BaselineSpline#Initialise()
//end

// initialize spline baseline package
static function Initialise()
	
	if (goodbye())
		return 0
	endif
	
	if (strlen(WinList("*",";","WIN:1")) == 0)  // no graphs
		DoAlert 0, "Spline baseline requires a trace plotted in a graph window."
		return 0
	endif
	
	// make sure the top graph is visible
	DoWindow /F $WinName(0,1)
	
	// clear any package detritus from the last used spline graph
	clearGraph(1)
	
	// initialize (or reinitialize) the package data folder
	DFREF dfr = resetDFR()
	wave /SDFR=dfr w_spline_dependency
	
	makePanel()
		
	string graphStr = getGraph()
	ControlInfo /W=$graphStr+"#SBL_panel" popTrace
	string traceStr = s_value
	
	if (kRecolorTraces==1)
		// save current trace colors and set all to grey
		resetColors(getGraph(), 3)
	endif
	
	variable success

	// place nodes within window
	success = resetNodes()
	if (!success)
		DoAlert 0, "Spline baseline requires a trace plotted in a graph window."
		clearGraph(1)
		return 0
	endif
	
	// set the sub and base waves for this trace
	success = resetGraph(traceStr)
	if (!success)
		DoAlert 0, "Spline baseline requires a trace plotted in a graph window."
		clearGraph(1)
		return 0
	endif
	
	string yAxisName = StringByKey("YAXIS", TraceInfo(graphStr,traceStr,0))
	ModifyGraph zero($yAxisName)=1
	
	// set a dependency to trigger interpolation and update graph when we
	// adjust a node position
	w_spline_dependency := BaselineSpline#updateSpline(root:Packages:SplineFit:w_nodesY)
	// won't work if we change package folder
	
	GraphWaveEdit /W=$graphStr/NI/T=1/M $"w_nodesY"
	// capture shift/ctrl/option-click
	SetWindow $graphStr hook(hSBL)=BaselineSpline#hookSplineGraph
		
	return 1
end

// uses the mac ctrl key, not command, for node insertion.
static function hookSplineGraph(STRUCT WMWinHookStruct &s)		
	if (s.eventcode!=5) // mouseup
		return 0
	endif
	
	variable v_key=GetKeyState(0)
	if (v_key == 0)
		return 0
	endif
	
	if (v_key&2^2) // shift click
		// toggle mode
		toggleEdit()
		return 0
	endif
	
	ControlInfo /W=$s.WinName+"#SBL_panel" check_edit
	if (v_value == 0)
		return 0
	endif

	variable insert=0
	insert = (stringmatch(IgorInfo(2),"Macintosh")) ? (v_key&2^4) : (v_key&2^0) // ctrl
	if (insert)
		DFREF dfr = getDFREF()
		wave /SDFR=dfr w_nodesY, w_nodesX
		variable v_X, v_Y
		
		string s_info = TraceInfo(s.WinName,"w_nodesY",0)
		string xAxisName = StringByKey("XAXIS", s_info)
		string yAxisName = StringByKey("YAXIS", s_info)
		
		v_X = AxisValFromPixel(s.WinName, xAxisName, s.mouseLoc.h)
		v_Y = AxisValFromPixel(s.WinName, yAxisName, s.mouseLoc.v)
		
		// deal with offset
		ControlInfo /W=$s.WinName+"#SBL_panel" popTrace
		string infoStr = TraceInfo(s.WinName,s_value,0)
		v_Y -= GetOffsetFromInfoString(infoStr, 1)
		
		w_nodesX[numpnts(w_nodesX)] = {v_X}
		w_nodesY[numpnts(w_nodesY)] = {v_Y}
		Sort w_nodesX, w_nodesX, w_nodesY
		updateSpline({NaN})
	endif

	return 0
end

static function toggleEdit()
	string graphStr = getGraph()
	ControlInfo /W=$graphStr+"#SBL_panel" check_edit
	setEdit(!v_value, graphStr)
end

static function setEdit(int v_edit, string graphStr)	
	if (v_edit == -1)
		ControlInfo /W=$graphStr+"#SBL_panel" check_edit
		v_edit = v_value
	endif	
	if (v_edit)
		// manually scale y axis
		string strYax = StringByKey("YAXIS", TraceInfo(graphStr,"w_nodesY",0))
		GetAxis /W=$graphStr/Q $strYax
		SetAxis /W=$graphStr $strYax, V_min, V_max
		
		ModifyGraph /W=$graphStr mode(w_nodesY)=3, marker(w_nodesY)=11
		GraphWaveEdit /W=$graphStr/NI/T=1/M $"w_nodesY"
		CheckBox check_edit win=$graphStr+"#SBL_panel", value=1
	else
		GraphNormal /W=$graphStr
		ModifyGraph /W=$graphStr mode(w_nodesY)=2
		CheckBox check_edit win=$graphStr+"#SBL_panel", value=0
	endif
end

static function resetGraph(string traceStr)		
	DFREF dfr = getDFREF()	
	string graphStr = getGraph()
	wave /Z w_data = TraceNameToWaveRef(graphStr, traceStr)
	if (WaveExists(w_data) == 0)
		return 0
	endif
	wave /Z w_x = XWaveRefFromTrace(graphStr, traceStr)
	
	if (WaveExists(w_x))
		// check for monotonic X wave
		Make /free w_diff
		Differentiate w_x /D=w_diff
		WaveStats /Q/M=0 w_diff
		if (V_max>0 && V_min<0)
			DoAlert 0, "X wave is not monatonic! Quitting :("
			clearGraph(1)
			return 0
		endif
	endif
	
	if (kRecolorTraces==1)
		resetColors(graphStr, 1) // set all to grey
		restoreTraceColor(graphStr, traceStr)
	endif
	
	string s_info = TraceInfo(graphStr,traceStr,0)
	string xAxisName = StringByKey("XAXIS", s_info)
	string yAxisName = StringByKey("YAXIS", s_info)
		
	// manually scale y axis
	GetAxis /Q $yAxisName
	SetAxis $yAxisName, V_min, V_max
		
	Duplicate /O w_data dfr:w_base, dfr:w_sub
	wave /SDFR=dfr w_nodesX, w_nodesY, w_base, w_sub
	
	// make sure base and sub waves are displayed properly
	appendToSameAxes(graphStr, traceStr, w_base, w_x, offset=1, unique=1)
	appendToSameAxes(graphStr, traceStr, w_sub, w_x, unique=1)
	appendToSameAxes(graphStr, traceStr, w_nodesY, w_nodesX, offset=1, unique=1)
	setEdit(-1, graphStr)	
		
	ModifyGraph /W=$graphStr live(w_base)=1
	ReorderTraces /W=$graphStr $"w_nodesY", {$"w_base"}

	return 1
end

static function selectTrace(string graphStr, string traceStr)
	if (kRecolorTraces == 1)
		graphStr = selectstring(strlen(graphStr)==0, graphStr, getGraph())
		resetColors(graphStr, 1) // set all to grey
		if (strlen(traceStr))
			restoreTraceColor(graphStr, traceStr)
		endif
		return 1
	endif
	return 0
end

static function resetNodes()
	
	string graphStr = getGraph()
	DFREF dfr = getDFREF()
	wave /SDFR=dfr w_nodesX, w_nodesY

	// define some local variables
	variable i, p_low, p_high, delP, delX, numNodes=5
	string infoStr
	variable trace_offset = 0
	
	ControlInfo /W=$graphStr+"#SBL_panel" popTrace
	string traceStr = s_value
	
	wave /Z w_data = TraceNameToWaveRef(graphStr, traceStr)
	wave /Z w_x = XWaveRefFromTrace(graphStr, traceStr)
	if (WaveExists(w_data) == 0)
		return 0
	endif
	Redimension /N=(numNodes) w_nodesX, w_nodesY
	
	// determine axis names
	string strInfo = TraceInfo(graphStr,traceStr,0)
	string xAxisName = StringByKey("XAXIS", strInfo)
	string yAxisName = StringByKey("YAXIS", strInfo)
	
	// determine where to place nodes on the bottom axis
	
	// keep nodes within window
	GetAxis/Q $xAxisName
	Make /N=2/free xrange={v_min, v_max}
	Sort xrange, xrange
	
	GetAxis/Q $yAxisName
	Make /N=2/free yrange={v_min, v_max}
	Sort yrange, yrange
	
	// keep nodes within extent of w_data
	if (WaveExists(w_x))
		xrange[0] = max(xrange[0],WaveMin(w_x))
		xrange[1] = min(xrange[1],WaveMax(w_x))
	else
		xrange[0] = max(xrange[0], min(pnt2x(w_data, 0), pnt2x(w_data, numpnts(w_data)-1)) )
		xrange[1] = min(xrange[1], max(pnt2x(w_data, 0), pnt2x(w_data, numpnts(w_data)-1)) )
	endif
	
	// inset nodes a little
	delX = (xrange[1] - xrange[0]) / 50
	xrange[0] += delX
	xrange[1] -= delX
	delX = (xrange[1] - xrange[0]) / (numNodes - 1)
	// initialize node positions
	w_nodesX = xrange[0] + p * delX
	
	int useAvg = kSetNodesAlgorithm == 0
	
	if (kSetNodesAlgorithm)
		useAvg = !AdvancedNodesGuess(w_data, w_x, w_nodesX, w_nodesY)
	endif
	
	if (useAvg)
		// the old way
		for (i=0;i<numNodes;i+=1)
			if (WaveExists(w_x))
				// necessary for unevenly spaced data
				FindLevel /Q/P w_x, w_nodesX[i]-delX/24
				if (v_flag)
					FindLevel /Q/P w_x, w_nodesX[i] // this HAS to be within range
				endif
				p_low = V_LevelX
				
				FindLevel /Q/P w_x, w_nodesX[i]+delX/24
				if (v_flag)
					FindLevel /Q/P w_x, w_nodesX[i] // this HAS to be within range
				endif
				p_high = V_LevelX
				WaveStats /Q/M=1/R=[p_low , p_high] /Z  w_data
			else
				WaveStats /Q/M=1/R=(w_nodesX[i]-delX/24 , w_nodesX[i]+delX/24) /Z  w_data
			endif
			w_nodesY[i] = V_avg  // this handles a few NANs
		endfor
	endif
	
	// Akima calculation won't work if nodes are reverse sorted
	Sort w_nodesX, w_nodesX, w_nodesY
	
	// if data are offset, then apply offsets to nodes
	infoStr = TraceInfo(graphStr,traceStr,0)
	sscanf ListMatch(infoStr, "offset(x)=*"), "offset(x)={%*g,%g}", trace_offset
	
	// keep nodes within window
	w_nodesY = max(yrange[0]-trace_offset, w_nodesY)
	w_nodesY = min(yrange[1]-trace_offset, w_nodesY)
	
	// make sure nodes are plotted on correct axes
	appendToSameAxes(graphStr, traceStr, w_nodesY, w_nodesX, offset=1, unique=1)
	setEdit(1, graphStr)
		
	return 1
end

// populate w_nodesY with guesses for a baseline
static function AdvancedNodesGuess(wave w_data, wave/Z w_x, wave w_nodesX, wave w_nodesY)
	duplicate /free/D w_data XforHull, YforHull
	if (waveexists(w_x))
		XforHull = w_x
	else
		XforHull = x
	endif
	// try to apply a bit of smoothing
	Loess /V=0/N=(max(7, numpnts(YforHull)/50))/E=1/Z srcWave=YforHull	
	// add concave function
	variable depth = (WaveMax(YforHull) - WaveMin(YforHull)) * .15
	variable radius = (WaveMax(XforHull) - WaveMin(XforHull)) / 2
	YforHull += depth * (XforHull[p] - XforHull[0])^2 / radius^2
	
	ConvexHull /Z XforHull, YforHull
	wave /Z w_XHull = w_XHull, w_YHull = w_YHull
	if (waveexists(w_XHull) == 0)
		return 0
	endif
	Reverse /P w_XHull, w_YHull
	WaveStats /M=1/Q w_XHull
	Rotate -v_minloc, w_XHull, w_YHull
	SetScale /P x, 0, 1, w_XHull, w_YHull
	WaveStats /M=1/Q w_XHull
	DeletePoints v_maxloc+1, numpnts(w_XHull)-v_maxloc-1, w_XHull, w_YHull
	w_YHull -= depth * (w_XHull[p] - XforHull[0])^2 / radius^2
	Interpolate2 /T=1/Y=w_nodesY/X=w_nodesX/I=3 W_XHull, W_YHull
	killWaves /Z w_XHull, w_YHull
	return 1
end

// place nodes on w_data.
// doesn't reset x positions unless they're out of range.
static function setNodes()
	
	string graphStr = getGraph()
	ControlInfo /W=$graphStr+"#SBL_panel" popTrace
	string traceStr = s_value
	wave /Z w_data = TraceNameToWaveRef(graphStr, traceStr)
	wave /Z w_x = XWaveRefFromTrace(graphStr, traceStr)
	
	DFREF dfr = getDFREF()
	wave /SDFR=dfr w_nodesX, w_nodesY
	variable numNodes = numpnts(w_nodesY)
	variable i, v_pnt, v_x
	
	for (i=0;i<numNodes;i+=1)
		if (WaveExists(w_x))
			// necessary for unevenly spaced data
			v_x = max(WaveMin(w_x), w_nodesX[i])
			v_x = min(WaveMax(w_x), v_x)
			
			FindLevel /Q/P w_x, v_x
			if (V_flag)
				v_pnt = numpnts(w_x)/2 // shouldn't arrive here
			else
				v_pnt = V_LevelX
			endif
			w_nodesY[i] = w_data[v_pnt]
		else
			// keep x values within range of w_data
			//	out of range error fixed with double precision nodes waves
			w_nodesX = max(min(pnt2x(w_data,0), pnt2x(w_data,numpnts(w_data)-1)), w_nodesX)
			w_nodesX = min(max(pnt2x(w_data,0), pnt2x(w_data,numpnts(w_data)-1)), w_nodesX)
			
			w_nodesY[i] = w_data(w_nodesX[i])
		endif
	endfor

	#if IgorVersion() >= 7
		// remove duplicate nodes,
		// in case some were outside of the range of w_data
		Duplicate /free w_nodesX w_duplicates
		FindDuplicates /SN=(NaN)/SNDS=w_duplicates w_nodesX
		i = 0
		do
			if (numtype(w_duplicates[i]) != 0)
				DeletePoints i, 1, w_nodesX, w_nodesY, w_duplicates
			else
				i += 1
			endif
		while(i < numpnts(w_nodesX))
	#else
		// you're out of luck
	#endif
	
	if (numpnts(w_nodesX) < 2)
		resetNodes()
		DoAlert 0, "Nodes couldn't be placed on " + NameOfWave(w_data) + " at the same x positions. Nodes have been reset."
	else
		if (kSetNodesAlgorithm)
			AdvancedNodesGuess(w_data, w_x, w_nodesX, w_nodesY)
		endif
		
		// make sure nodes are plotted on correct axes
		appendToSameAxes(graphStr, traceStr, w_nodesY, w_nodesX, offset=1, unique=1)
		setEdit(1, graphStr) // also sets mode for nodes trace
	endif
end

static function makePanel()		
	DFREF dfr = getDFREF()	
	// make sure the top graph is visible
	string graphStr = WinName(0,1)
	DoWindow /F $graphStr
	
	// make sure there's space to the left of the graph for the panel
	GetWindow /Z $graphStr wsize
	if (V_left<190)
	#if IgorVersion() >= 7
		MoveWindow /W=$graphStr 190, V_top, -1, -1
	#else
		MoveWindow /W=$graphStr 190, V_top, 190+(V_right-V_left), v_bottom
	#endif
	endif

	// make panel
	NewPanel /K=1/N=SBL_panel/W=(190,0,0,380)/HOST=$graphStr/EXT=1 as "Spline Controls"
	ModifyPanel /W=$graphStr#SBL_panel, noEdit=1
	
	string traceStr = StringFromList(0, splineTraces())
	wave /Z w = TraceNameToWaveRef(graphStr, traceStr)
	
	variable i = 0, deltaY = 25, font = 12, groupw = 180
	v_left = 25
	v_top = 5

	GroupBox group0,pos={v_left-20,v_top+deltaY*i},size={groupw,deltaY*2},title="Data wave"
	GroupBox group0,fSize=font
	i += 1
	// store values internally in these controls
	PopupMenu popTrace, mode=1, Value=BaselineSpline#splineTraces(), title="",pos={v_left,v_top+deltaY*i},size={130,20}
	PopupMenu popTrace, help={"select data wave" }, Proc=BaselineSpline#popupTrace, fSize=font
	i += 1.5
	
	GroupBox group1 pos={v_left-20,v_top+deltaY*i},size={groupw,deltaY*6.0},title="Spline parameters"
	GroupBox group1 fSize=font
	i += 1
	CheckBox check_akima, pos={v_left,v_top+deltaY*i}, fSize=font, value=0, Proc=BaselineSpline#splineCheckboxes, title="Akima"
	CheckBox check_akima, mode=1,help={"use Akima spline"}
	i += 0.8
	CheckBox check_PCHIP, pos={v_left,v_top+deltaY*i}, fSize=font, value=0, Proc=BaselineSpline#splineCheckboxes, title="PCHIP"
	CheckBox check_PCHIP, mode=1,help={"use piecewise cubic Hermite spline"}
	i += 0.8
	CheckBox check_cubic, pos={v_left,v_top+deltaY*i}, fSize=font, value=1, Proc=BaselineSpline#splineCheckboxes, title="Cubic"
	CheckBox check_cubic, mode=1,help={"use cubic spline"}
	i += 0.8
	PopupMenu popE,pos={v_left+35,v_top+deltaY*i},size={46,21}, fSize=font, title="Ends:", Proc=BaselineSpline#popupEnds
	PopupMenu popE, value="1st deriv.;natural", mode=2
	i += 0.8
	CheckBox check_linear, pos={v_left,v_top+deltaY*i}, fSize=font, value=0, Proc=BaselineSpline#splineCheckboxes, title="Linear"
	CheckBox check_linear, mode=1,help={"lines between nodes"}
	i += 0.8
	CheckBox check_smoothing, pos={v_left,v_top+deltaY*i}, fSize=font, value=0, Proc=BaselineSpline#splineCheckboxes, title="Smoothing"
	CheckBox check_smoothing, mode=1,help={"use smoothing spline"}
	SetVariable setvarF, pos={v_left+100,v_top+deltaY*i},size={50,16},title="",bodyWidth=50, focusring=0
	SetVariable setvarF, value=_NUM:1, limits={0,Inf,1}, fSize=font, Proc=BaselineSpline#splineSetvars
	i += 1.5
	
	GroupBox group2,pos={v_left-20,v_top+deltaY*i},size={groupw,deltaY*4.5},title="Node control"
	GroupBox group2,fSize=font
	i += 1
	CheckBox check_edit, pos={v_left+30,v_top+deltaY*i}, fSize=font, value=1, Proc=BaselineSpline#splineCheckboxes, title="Edit mode "
	CheckBox check_edit, mode=0,help={"Toggle to edit nodes"}, side=1
	i += 0.8
	SetDrawLayer ProgBack
	SetDrawEnv textyjust=2
	DrawText v_left-12,v_top+deltaY*i,"Shift-click to toggle mode, ctrl-\r/option-click to add/zap nodes"
	i += 1.5
	Button btnResetNodes,pos={v_left,v_top+deltaY*i},size={50,20},title="Reset", Proc=BaselineSpline#splineButtons
	Button btnResetNodes,help={"Distribute nodes over x-range of graph"}, fSize=font
	
	Button btnLoadNodes,pos={v_left+80,v_top+deltaY*i},size={50,20},title="Load...", Proc=BaselineSpline#splineButtons
	Button btnLoadNodes,help={"Load nodes used in a previous fit"} , fSize=font
	
	i += 1.5
	Button SBL_sub,pos={60,v_top+deltaY*i},size={70,20},title="Subtract", Proc=BaselineSpline#splineButtons
	Button SBL_sub,help={"Subtract baseline and save a copy"}, fSize=font
	
	return 1
end

static function splineSetvars(STRUCT WMSetVariableAction &s)	
	if (s.eventCode<0)
		return 0
	endif
	updateSpline({NaN})
end

static function popupEnds(STRUCT WMPopupAction &s)
	if (s.eventCode == 2)
		updateSpline({NaN})
	endif
end

static function popupTrace(STRUCT WMPopupAction &s)	
	switch (s.eventCode)
		case 3 :
			if (igorversion() >= 9)
				string graphStr = ParseFilePath(0, s.win, "#", 0, 0)
				selectTrace(graphStr, s.popStr)
			endif
			break
		case 2:
			ResetGraph(s.popStr)
			setNodes()
			updateSpline({NaN})
			break 
		case 4: 	
			ResetGraph(s.popStr)
			updateSpline({NaN})
			break
	endswitch
	return 0		
end
  
static function splineCheckboxes(STRUCT WMCheckboxAction &s)	
	if (s.eventCode == -1)
		clearGraph(0)
		return 0
	elseif (s.eventCode != 2)
		return 0
	endif
	
	string checkboxList = "check_akima;check_PCHIP;check_cubic;check_linear;check_smoothing"
	int i
	int numTypes = ItemsInList(checkboxList)
	int checkboxNum = WhichListItem(s.ctrlName, checkboxList)

	if (checkboxNum>=0)
		for (i=0;i<numTypes;i+=1)
			if (i != checkboxNum)
				CheckBox $StringFromList(i, checkboxList), win=$s.win, value=0
			endif
		endfor
		updateSpline({NaN})
		return 0
	endif
	
	// only remaining checkbox is check_edit
	string graphStr = ParseFilePath(0, s.win, "#", 0, 0)
	setEdit(s.checked, graphStr)

	return 1
end

static function splineButtons(STRUCT WMButtonAction &s)	
	if (s.eventCode != 2)
		return 0
	endif	
	if (stringmatch(s.ctrlname, "SBL_sub"))
		subtractBaseline()
	elseif (stringmatch(s.ctrlname, "btnResetNodes"))
		resetNodes()
	elseif (stringmatch(s.ctrlname, "btnLoadNodes"))
		loadNodes()
	endif
	return 0
end

static function /S splineTraces()	
	string graphStr = getGraph()
	string listStr = TraceNameList(graphStr,";",1+4)
	string removeStr = "w_base;w_NodesY;"
	removeStr += ListMatch(listStr, "*" + ksSubtractedSuffix)
	removeStr += ListMatch(listStr, "*" + ksBaselineSuffix)
	listStr = RemoveFromList(removeStr, listStr, ";", 0)
	return listStr
end

// initialize (or reinitialize) the package data folder
static function /DF resetDFR()	
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:SplineFit
	DFREF dfr = root:Packages:SplineFit
	wave /Z w_nodesX = dfr:w_nodesX
	if (WaveType(w_nodesX) != 4)
		Make /D/O/N=5 dfr:w_nodesX, dfr:w_nodesY
	endif
	Make /O/N=1 dfr:w_spline_dependency={NaN}
	return dfr
end

static function/DF getDFREF()
	DFREF dfr = root:Packages:SplineFit
	if (DataFolderRefStatus(dfr) != 1)
		DFREF dfr = resetDFR()
	endif
	return dfr
end

// dependency function to define the spline curve
static function updateSpline(wave w)
		
	DFREF dfr = getDFREF()
	wave /Z/SDFR=dfr w_nodesX,w_nodesY,w_base,w_sub
	
	string graphStr = getGraph()
	if (strlen(graphStr) == 0)
		return 0
	endif
	string panelStr = graphStr+"#SBL_panel"
		
	ControlInfo /W=$panelStr popTrace
	string traceStr = s_value
	wave /Z w_data = TraceNameToWaveRef(graphStr, traceStr)
	wave /Z w_x = XWaveRefFromTrace(graphStr, traceStr)
	
	if (WaveExists(w_data) == 0)
		return NaN
	elseif (numpnts(w_nodesX) < 2)
		return NaN
	elseif (numpnts(w_data) != numpnts(w_base))
		// could get here if we trigger an update before popup runs updateGraph
		return NaN
	endif
	
	// flags for interpolate2
	variable flagE, flagF, type=1 // set default to linear interpolation
	// if type < 1, spline is calculated using a method other than Igor's interpolate2
	
	ControlInfo /W=$panelStr check_akima
	if (v_value)
		type = 0
	endif
	
	ControlInfo /W=$panelStr check_PCHIP
	if (v_value)
		type = -1
	endif
	
	if (type == 0 && numpnts(w_nodesX)<5)
		DoAlert 0, "Akima spline requires at least 5 nodes\r - reverting to standard cubic spline"
		CheckBox check_akima, win=$panelStr, value=0
		CheckBox check_cubic, win=$panelStr, value=1
		type = 2
	endif
		
	// check for small number of points
	if (type>0 && numpnts(w_nodesX)<4)
		type = 1 // don't uncheck spline type, so will revert to desired type as nodes are added
	endif
		
	if (type>0) // types that use interpolate2
		do
			// collect info to set the flags for interpolate2
			ControlInfo /W=$panelStr popE
			flagE = V_Value
			ControlInfo /W=$panelStr check_cubic
			if (v_value)
				type = 2
				break
			endif
			ControlInfo /W=$panelStr check_smoothing
			if (v_value)
				type = 3
				ControlInfo /W=$panelStr setvarF
				flagF = V_value
				break
			endif
		while(0)
	endif

	// do the spline calculation
	if (type == -1) // PCHIP
		PCHIP(w_nodesX, w_nodesY, w_base, xwave=w_x)
	elseif (type == 0) // Akima type spline
		Akima(w_nodesX, w_nodesY, w_base, xwave=w_x)
	elseif (type > 0) // cubic spline
		DebuggerOptions
		variable sav_debug = V_debugOnError
		DebuggerOptions debugOnError = 0 // switch this off in case interpolate fails
		try
			if (WaveExists(w_x))
				Interpolate2 /E=(flagE)/F=(flagF)/T=(type)/I=3/X=w_x/Y=w_base w_nodesX, w_nodesY; AbortOnRTE
			else
				Interpolate2 /E=(flagE)/F=(flagF)/T=(type)/I=3/Y=w_base w_nodesX, w_nodesY; AbortOnRTE
			endif
		catch
			Print "Error during interpolate:"
			variable CFerror = GetRTError(1)
			Print GetErrMessage(CFerror)
		endtry
		DebuggerOptions debugOnError=sav_debug
	endif
	
	w_sub = w_data - w_base
	
	return NaN
end

// function to subtract the spline curve
static function subtractBaseline()	
	string graphStr = getGraph()
	string panelStr = graphStr + "#SBL_panel"
	ControlInfo /W=$panelStr popTrace
	string traceStr = s_value
	wave /Z w_data = TraceNameToWaveRef(graphStr, traceStr)
	wave /Z w_x = XWaveRefFromTrace(graphStr, traceStr)
	
	if (WaveExists(w_data) == 0)
		return NaN
	endif
	
	DFREF dfr = getDFREF()
	wave /SDFR=dfr w_nodesX, w_nodesY, w_base
	
	string strSubtracted, strBaseline, strNodes
	strSubtracted = CleanupName(NameOfWave(w_data) + ksSubtractedSuffix, 0)
	strBaseline = CleanupName(NameOfWave(w_data) + ksBaselineSuffix, 0)
	strNodes = CleanupName(NameOfWave(w_data)+"_SpNodes", 0)
	
	DFREF currentDFR = GetDataFolderDFR()
	
	if (exists(strSubtracted))
		DoAlert 1, strSubtracted + " exists. Overwrite?"
		if (V_flag == 2)
			return 0
		endif
	endif
	Duplicate /O w_data currentDFR:$strSubtracted /WAVE=nob
	
	if (exists(strBaseline))
		DoAlert 1, strBaseline + " exists. Overwrite?"
		if (V_flag == 2)
			return 0
		endif
	endif
	Duplicate /O w_base currentDFR:$strBaseline /WAVE=bl
	
	if (exists(strNodes))
		DoAlert 1, strNodes + " exists. Overwrite?"
		if (V_flag == 2)
			return 0
		endif
	endif
	Duplicate /O w_nodesX currentDFR:$strNodes /WAVE=nd
	Redimension /N=(-1, 2) nd
	nd[][0] = w_nodesX[p]
	nd[][1] = w_nodesY[p]
	
	nob = w_data - bl
	
	string noteStr="Baseline Parameters:"
	ControlInfo /W=$panelStr check_akima
	if (v_value)
		noteStr += "Type=AkimaSpline;"
	endif
	ControlInfo /W=$panelStr check_cubic
	if (v_value)
		noteStr += "Type=CubicSpline;"
		ControlInfo /W=$panelStr popE
		noteStr += "Ends="+s_value+";"
	endif
	ControlInfo /W=$panelStr check_PCHIP
	if (v_value)
		noteStr += "Type=PCHIP;"
	endif
	ControlInfo /W=$panelStr check_smoothing
	if (v_value)
		noteStr += "Type=SmoothingSpline;"
		ControlInfo /W=$panelStr setvarF
		noteStr += "SmoothingFactor="+num2str(v_value)+";"
	endif
	ControlInfo /W=$panelStr check_linear
	if (v_value)
		noteStr += "Type=LinearSpline;"
	endif
	noteStr += "numNodes="+num2str(numpnts(w_nodesX))+";"
	noteStr += "Data="+NameOfWave(w_data)+";Output="+NameOfWave(nob)+","+NameOfWave(bl)+","+NameOfWave(nd)+";\r"
	note /k bl, noteStr
	note nob, noteStr // append note from baseline wave to output wave note
	printf noteStr
		
	// add baseline-subtracted output wave to plot
	appendToSameAxes(graphStr, traceStr, nob, w_x, matchRGB=1, unique=1)
	
	// possibly plot baseline
	if (kAppendBaselines)
		appendToSameAxes(graphStr, traceStr, bl, w_x, matchRGB=1, unique=1)
	endif
	
	if (kRemoveOriginal)
		RemoveFromGraph /W=$graphStr/Z $traceStr, w_base
		PopupMenu popTrace, win=$panelStr, mode=1
		ResetGraph(getSelectedTrace())
		setNodes()
		updateSpline({NaN})
	endif
	
	setEdit(0, graphStr)
	return 1
end

// function to clear the spline curve
static function clearGraph(int killPanel)	
	string graphStr = getGraph()
	
	if (strlen(graphStr))
		GraphNormal /W=$graphStr
		RemoveFromGraph /W=$graphStr/Z w_base, w_nodesY, w_sub
		SetWindow $graphStr hook(hSBL)=$""
	endif
	
	DFREF dfr = getDFREF()
	wave /Z w = dfr:w_spline_dependency
	if (WaveExists(w))
		SetFormula w, "" // remove dependency
	endif
	
	if (killPanel && strlen(graphStr))
		KillWindow $graphStr + "#SBL_panel"
	endif
	
	if (DataFolderRefStatus(dfr)>0 && kRecolorTraces==1)
		resetColors(graphStr, 0)
	endif
	
	return 1
end

static function loadNodes()
	DFREF dfr = getDFREF()
			
	string str_NodeWave
	variable LoadY
	
	Prompt str_NodeWave, "Select saved nodes", Popup, WaveList("*",";","DIMS:2")
	Prompt LoadY, " ", Popup, "Load X positions only;Load X and Y positions;"
	DoPrompt "", str_NodeWave, LoadY
	if (V_Flag)
		return 0
	endif
	
	wave w_nodes = $str_NodeWave
	loadY -= 1
	
	wave /SDFR=dfr w_nodesX, w_nodesY
	
	Redimension /N=(DimSize(w_nodes, 0)), w_nodesX, w_nodesY
	w_nodesX = w_nodes[p][0]
	
	if (LoadY)
		w_nodesY = w_nodes[p][1]
		return 1
	endif
	
	// set the y values to follow data wave
	setNodes()

	return 1
end


// ---------------------- Akima spline code --------------------------

// this part based on code posted to IgorExchange by Michael Bongard
// edited to allow extrapolation beyond end nodes
 
// Akima.ipf: Routines to implement Akima-spline fitting, based on
// H. Akima, JACM, Vol 17, No 4, 1970 p 589-602
// after M. Bongard, 11/17/09

static function Akima(wave w_nodesX, wave w_nodesY, wave w_interp, [wave/Z xwave])	
	wave w_iota = AkimaIota(w_nodesX, w_nodesY)
	if (WaveExists(w_iota))
		if (WaveExists(xwave))
			multithread w_interp = AkimaThread(xwave[p], w_nodesX, w_nodesY, w_iota)
		else
			multithread w_interp = AkimaThread(x, w_nodesX, w_nodesY, w_iota)
		endif
		return 1
	endif
	return 0
end

static function /wave AkimaIota(wave knotX, wave knotY)
	
 	// removed some sanity checks, converted to wave function
	variable numKnots = numpnts(knotX)
 
	if (numKnots < 5)
		Print "calcIota: ERROR -- Akima spline algorithm requires at least 5 knots. Aborting..."
		return $""
	endif
 	 
	Make/D/FREE/N=(numKnots) kX, kY
	kX = knotX
	kY = knotY
 
	// Handle end-point extrapolation
	Make/D/FREE/N=5 endX, endY
	
	// RHS: end points are last three in dataset
	endX[0,2] = kX[p+numKnots-3]
	endY[0,2] = kY[p+numKnots-3]
 
	AkimaEndPoints(endX, endY)
 	
 	kX[numpnts(kX)]={endX[3],endX[4]}
 	kY[numpnts(kY)]={endY[3],endY[4]}

	// LHS: end points are first three in dataset, but reversed in ordering
	// (i.e. point 3 in Akima's notation == index 0)
	endX = 0
	endY = 0
	
	endX[0,2] = knotX[2-p]
	endY[0,2] = knotY[2-p]
 
	AkimaEndPoints(endX, endY)
	
	InsertPoints  0, 2, kX, kY
	kX[0,1] = endX[4-p]
	kY[0,1] = endY[4-p]
 
	// kX, kY are now properly populated, including extrapolated endpoints
 
	Make /D/FREE/N=(numKnots + 4 - 1) mK
	mK = (kY[p+1]-kY)/(kX[p+1]-kX)
 
	Make /free/N=(numKnots) w_Iota
	w_Iota = ( abs(mK[p+3]-mK[p+2])*mK[p+1] + abs(mK[p+1]-mK[p])*mK[p+2] ) / (abs(mK[p+3]-mK[p+2])+abs(mK[p+1]-mK))
	w_Iota = numtype(w_Iota) == 0 ? w_Iota : 0.5*(mK[p+1]+mK[p+2])
	return w_Iota
end

// Given: 5-point knot wave knotX, knotY, with i=[0,2] representing the last three
// knot locations from data, compute end knots i=[3,4] appropriately.
// kX and kY hold node X, Y coordinates, respectively
threadsafe static function AkimaEndPoints(wave kX, wave kY) 
 
	// compute X locations of knots, eq. 8 of Akima 1970:
	kX[3] = kX[1] + kX[2] - kX[0]
	kX[4] = 2*kX[2] - kX[0]
 
	// eq. (12)-(14), determine gradient on [0,1]
	Make /N=4/FREE mi
	mi = (kY[p+1]-kY)/(kX[p+1]-kX)
 
	// Determine remainder of quantities by applying solutions of eq (9)
	kY[3] = (2*mi[1] - mi[0])*(kX[3] - kX[2]) + kY[2]
	mi[2] = (kY[3] - kY[2]) / (kX[3] - kX[2])
 
	kY[4] = (2*mi[2] - mi[1])*(kX[4] - kX[3]) + kY[3]
	// not sure why the next line was included
	//mi[3] = (kY[4] - kY[3]) / (kX[4] - kX[3])
end

threadsafe static function AkimaThread(variable x, wave w_nodesX, wave w_nodesY, wave w_Iota) 	
	variable p0, p1
	if (x < WaveMin(w_nodesX))
		p0 = 0
	elseif (x >= WaveMax(w_nodesX))
		p0 = numpnts(w_nodesX) - 2
	else
		FindLevel /P/EDGE=1/Q w_nodesX, x
		if (v_flag == 1) // not found
			return NaN
		endif
		p0 = floor(v_levelX)
	endif
	p1 = p0 + 1

	variable x1, x2, y1, y2, iota1, iota2
	x1 = w_nodesX[p0]
	y1 = w_nodesY[p0]
	iota1 = w_Iota[p0]
 
	x2 = w_nodesX[p1]
	y2 = w_nodesY[p1]
	iota2 = w_Iota[p1]
 
	variable coeff2, coeff3, delx
	coeff2 = ( 3 * (y2 - y1)/(x2 - x1) - 2*iota1 - iota2 ) / (x2 - x1)
	coeff3 = (iota1 + iota2 - 2 * (y2 - y1)/(x2 - x1)) / (x2 - x1)^2
	delx = x - x1
 
	return y1 + iota1*delx + coeff2*delx^2 + coeff3*delx^3
end

// ----------------------------- PCHIP ---------------------------

// populate w_interp with interpolated values from piecewise cubic
// Hermite spline
static function PCHIP(wave w_nodesX, wave w_nodesY, wave w_interp, [wave /Z xwave])	
	if (numpnts(w_nodesX) == 2) // linear interpolation		
		w_interp = w_nodesY[0] + ((WaveExists(xwave) ? xwave[p] : x) - w_nodesX[0]) / (w_nodesX[1]-w_nodesX[0])*(w_nodesY[1]-w_nodesY[0])		
		return 1
	endif	
	wave w_d = setDerivativesPCHIP(w_nodesX, w_nodesY)
	w_interp = InterpolatePCHIP(w_nodesX, w_nodesY, w_d, WaveExists(xwave) ? xwave[p] : x)	
end

// The trick is to choose good values for the slope at each node. See:
// FN Fritsch and RE Carlson (1980) Monotone piecewise cubic
// interpolation. SIAM J. Numer. Anal. 17: 238. DOI:10.1137/0717021
// KW Brodlie (1980) A review of methods for curve and function drawing.
// In Mathematical Methods in Computer Graphics and Design, KW Brodlie,
// ed., Academic Press, London, pp. 1-37.
// FN Fritsch and J Butland (1984) A method for constructing local
// monotone piecewise cubic interpolants, SIAM Journal on Scientific and
// Statistical Computing 5: 300-304.
static function /WAVE setDerivativesPCHIP(wave w_nodesX, wave w_nodesY)
	Duplicate /free w_nodesX, w_d, w_a
	// w_d will be set to desired derivatives at node positions
	// w_a will be used as weighting for harmonic means
	w_d = 0
	
	Make /free/N=(numpnts(w_nodesX)-1), w_m, w_delx
	// w_m[i] is gradient between node i and i+1
	// w_delx[i] and w_delx[i-1] are x offsets of nodes i+1 and i-1
	
	w_m = (w_nodesY[p+1]-w_nodesY) / (w_nodesX[p+1]-w_nodesX)
	w_delx = w_nodesX[p+1] - w_nodesX
	variable pmax = numpnts(w_d) - 2  // pmax is numpnts(w_m)-1 = numpnts(w_d)-2
	
	w_a[1,pmax] = ( 1 + w_delx[p-1] / (w_delx[p-1] + w_delx) ) / 3
	
	// Brodlie modification of Butland formula (for same sign slopes)
	w_d [1,pmax] = ( w_m!=0 && w_m[p-1]!=0 && (sign(w_m)==sign(w_m[p-1])) ) ? w_m[p-1]*w_m/(w_a*w_m[p-1]+(1-w_a)*w_m) : 0
		
	// deal with ends
	w_d[0] = EndDerivativePCHIP(w_delx[0], w_delx[1], w_m[0], w_m[1])
	w_d[pmax+1] = EndDerivativePCHIP(w_delx[pmax], w_delx[pmax-1], w_m[pmax], w_m[pmax-1])
	
	// clean up any division by zero errors
	w_d = numtype(w_d) == 0 ? w_d : 0
			
	return w_d
end

// one-sided three-point estimate for the derivative adjusted to be shape
// preserving
static function endDerivativePCHIP(variable delX0, variable delX1, variable m0, variable m1)
	variable derivative = ( m0*(2*delX0+delX1) - delX0*m1 ) / (delX0+delX1)
	if (sign(derivative) != sign(m0))
		derivative = 0
	elseif ( (sign(m0)!=sign(m1)) && (abs(derivative)>3*abs(m0)) )
		derivative = 3 * m0
	endif
	return derivative
end

threadsafe static function interpolatePCHIP(wave w_nodesX, wave w_nodesY, wave w_nodesM, variable x)
	variable p0, p1
	if (x < WaveMin(w_nodesX))
		p0 = 0
	elseif (x >= WaveMax(w_nodesX))
		p0 = numpnts(w_nodesX) - 2
	else
		FindLevel /P/EDGE=1/Q w_nodesX, x
		if (v_flag == 1) // not found
			return NaN
		endif
		p0 = floor(v_levelX)
	endif	
	p1 = p0 + 1
	return InterpolateHermite(x, w_nodesX[p0], w_nodesY[p0], w_nodesX[p1], w_nodesY[p1], w_nodesM[p0], w_nodesM[p1])
end

// https: en.wikipedia.org/wiki/Cubic_Hermite_spline
// calculate cubic function f(x) such that f(x1)=y1, f(x2)=y2, f'(x1)=m1 and f'(x2)=m2
threadsafe static function interpolateHermite(variable x, variable x0, variable y0, variable x1, variable y1, variable m0, variable m1)
	variable t = (x-x0) / (x1-x0)
	m0 *= (x1-x0)
	m1 *= (x1-x0)
	return y0*(2*t^3-3*t^2+1) + m0*(t^3-2*t^2+t) + y1*(-2*t^3+3*t^2) + m1*(t^3-t^2)
end

// ----------------------------------------------------------------------------

// Plots w, optionally vs w_x, on the same axes as the already plotted
// trace 'traceStr'
// appendToSameAxes(graphStr, traceStr, w, w_x, w_rgb={r,g,b}, offset=1)
// appends w (vs w_x if w_x exists), sets color to (r,g,b) and matches y
// offset of traceStr
// Default is choose a color that contrasts with that of traceStr; matchRGB=1 forces
// color to match
// fill = val adds fill to zero with opacity=val and sends trace to back
// unique = 1 removes any other instances of wave w from the graph
// replace = 1 removes tracestr after plotting w
static function appendToSameAxes(graphStr, traceStr, w, w_x, [w_rgb, offset, matchRGB, fill, lsize, unique, replace])
	string graphStr, traceStr
	wave /Z w, w_x, w_rgb
	variable offset  // match y offset
	int matchRGB // match color of already plotted trace
	variable fill // 10% opacity fill to zero
	variable lsize
	int unique, replace
	
	offset = ParamIsDefault(offset) ? 0 : offset
	matchRGB = ParamIsDefault(matchRGB) ? 0 : matchRGB
	fill = ParamIsDefault(fill) ? 0 : min(fill, 1)
	lsize = ParamIsDefault(lsize) ? 1 : lsize
	unique = ParamIsDefault(unique) ? 0 : unique
	replace = ParamIsDefault(replace) ? 0 : replace
	
	string s_info = TraceInfo(graphStr, traceStr, 0)
	string s_Xax = StringByKey("XAXIS",s_info)
	string s_Yax = StringByKey("YAXIS",s_info)
	string s_flags = StringByKey("AXISFLAGS",s_info)
	variable flagBits = GrepString(s_flags, "/R") + 2*GrepString(s_flags, "/T")
	offset = offset ? GetOffsetFromInfoString(s_info, 1) : 0
	
	// get color of already plotted trace
	variable c0, c1, c2
	sscanf ListMatch(s_info, "rgb(x)=*"), "rgb(x)=(%d,%d,%d", c0, c1, c2
	
	if (matchRGB==0 && ParamIsDefault(w_rgb)) // no color specified
		STRUCT RGBcolor RGB
		RGB.red = c0; RGB.green = c1; RGB.blue = c2
		contrastingColor(RGB)
		Make /free w_RGB = {RGB.red, RGB.green, RGB.blue}
	endif
	if (matchRGB) // this overides any specified color
		Make /free/O w_RGB = {c0,c1,c2}
	endif
	
	int i, numTraces
	string traces, trace
	if (unique)
		traces = TraceNameList(graphStr,";",1)
		for (i=ItemsInList(traces)-1;i>=0;i-=1)
			trace = StringFromList(i,traces)
			wave /Z ithTraceWave = TraceNameToWaveRef(graphStr, trace)
			if (WaveRefsEqual(ithTraceWave, w))
				RemoveFromGraph /W=$graphStr/Z $trace
			endif
		endfor
	endif
	
	switch (flagBits)
		case 0:
			if (WaveExists(w_x))
				AppendToGraph /W=$graphStr/B=$s_Xax/L=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w vs w_x
			else
				AppendToGraph /W=$graphStr/B=$s_Xax/L=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w
			endif
			break
		case 1:
			if (WaveExists(w_x))
				AppendToGraph /W=$graphStr/B=$s_Xax/R=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w vs w_x
			else
				AppendToGraph /W=$graphStr/B=$s_Xax/R=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w
			endif
			break
		case 2:
			if (WaveExists(w_x))
				AppendToGraph /W=$graphStr/T=$s_Xax/L=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w vs w_x
			else
				AppendToGraph /W=$graphStr/T=$s_Xax/L=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w
			endif
			break
		case 3:
			if (WaveExists(w_x))
				AppendToGraph /W=$graphStr/T=$s_Xax/R=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w vs w_x
			else
				AppendToGraph /W=$graphStr/T=$s_Xax/R=$s_Yax/C=(w_rgb[0],w_rgb[1],w_rgb[2]) w
			endif
			break
	endswitch
	ModifyGraph /W=$graphStr offset($NameOfWave(w))={0,offset}, lsize($NameOfWave(w))=lsize
	
	if (fill > 0)
		ModifyGraph /W=$graphStr mode($NameOfWave(w))=7, hbFill($NameOfWave(w))=2
		ModifyGraph /W=$graphStr usePlusRGB($NameOfWave(w))=1, plusRGB($NameOfWave(w))=(w_rgb[0],w_rgb[1],w_rgb[2],fill*65535)
		ModifyGraph /W=$graphStr useNegRGB($NameOfWave(w))=1, negRGB($NameOfWave(w))=(w_rgb[0],w_rgb[1],w_rgb[2],fill*65535)
		ReorderTraces /W=$graphStr _back_, {$NameOfWave(w)}
	endif
	
	if (replace)
		RemoveFromGraph /W=$graphStr traceStr
	endif
end

// axis: 0 = x, 1 = y
static function GetOffsetFromInfoString(string s, int axis)
	variable xOffset, yOffset
	sscanf ListMatch(s, "offset(x)=*"), "offset(x)={%g,%g}", xOffset, yOffset
	return axis ? yOffset : xOffset
end

static function contrastingColor(STRUCT RGBcolor &RGB)
	STRUCT HSLcolor HSL
	RGB2HSL(RGB, HSL)
	if (HSL.S<0.3)
		RGB.red = 65535;RGB.green = 0;RGB.blue = 0
		return 1
	endif
	variable hue = round(6*HSL.H)
	hue -= 6*(hue>5)
	switch (hue)
		case 0: // red
			RGB.red = 0; RGB.green = 65535; RGB.blue = 65535 // cyan
			break
		case 1: // yellow
			RGB.red = 0; RGB.green = 0; RGB.blue = 65535 // blue
			break
		case 2: // green
			RGB.red = 65535; RGB.green = 0; RGB.blue = 65535 // magenta
			break
		case 3: // cyan
			RGB.red = 65535; RGB.green = 0; RGB.blue = 0 // red
			break
		case 4: // blue
//			RGB.red = 65535; RGB.green = 65535; RGB.blue = 0 // yellow
			RGB.red = 65535; RGB.green = 43690; RGB.blue = 0 // orange, yellow is hard to see
			break
		case 5: // magenta
			RGB.red = 0; RGB.green = 65535; RGB.blue = 0 // lime
			break
	endswitch
	return 1
end

// see easyrgb.com
static function RGB2HSL(STRUCT RGBcolor &rgb, STRUCT HSLcolor &hsl)
	// R, G and B input range = 0 - 65535
	// H, S and L output range = 0 - 1
	variable red = rgb.red/65535
	variable green = rgb.green/65535
	variable blue = rgb.blue/65535

	variable var_Min = min(red,green,blue)
	variable var_Max = max(red,green,blue)
	variable del_Max = var_Max - var_Min
	
	variable H,S,L
	L = (var_Max + var_Min)/ 2
	if (del_Max == 0) // grey
	    H = 0
	    S = 0
	else // Chromatic data
		S = (L < 0.5) ? del_Max/(var_Max + var_Min) : del_Max/(2 - var_Max - var_Min)
		
		variable del_R = ( (var_Max - red)/6 + del_Max/2 ) / del_Max
		variable del_G = ( (var_Max - green)/6 + del_Max/2 ) / del_Max
		variable del_B = ( (var_Max - blue)/6 + del_Max/2 ) / del_Max
		
		if (red == var_Max )
			H = del_B - del_G
		elseif (green == var_Max )
			H = (1/3) + del_R - del_B
		elseif (blue == var_Max )
			H = (2/3) + del_G - del_R
		endif
		H += (H < 0)
		H -= (H > 1)
	endif
	hsl.H = H
	hsl.S = S
	hsl.L = L
	return 1
end

static structure HSLcolor
	float H
	float S
	float L
endstructure

// ----------------------------------------------------------------------------
 
static function /T getGraph()
	GetWindow /Z SBL_panel activeSW
	return ParseFilePath(0, s_value, "#", 0, 0)
end

static function /T getSelectedTrace()
	ControlInfo /W=$getGraph()+"#SBL_panel" popTrace
	return s_value	
end

// status 0: reset to saved colors
// status 1: set all to grey
// status 2: save current
// status 3: save current and set all to grey
static function ResetColors(graphStr, status)
	string graphStr
	variable status
		
	DFREF dfr = GetDFREF()
	string traceStr
	variable i, numTraces
		
	if (status==0)
		wave /Z/T w_traces = dfr:w_traces
		wave /Z w_colors = dfr:w_colors
		if (WaveExists(w_colors)==0)
			return 0
		endif
		numTraces = numpnts(w_traces)
		for(i=0;i<numTraces;i+=1)
			ModifyGraph /W=$graphStr/Z rgb($w_traces[i])=(w_colors[i][0],w_colors[i][1],w_colors[i][2])
		endfor
		KillWaves /Z w_colors, w_traces
		return 1
	endif
	
	string traceListStr = TraceNameList(graphStr, ";", 1+4)
	numTraces = ItemsInList(traceListStr)
		
	if (status & 2)
		Make /O/N=0/T dfr:w_traces /wave=w_traces
		Make /O/N=(0,3) dfr:w_colors /wave=w_colors
	else
		wave /Z/T w_traces = dfr:w_traces
		wave /Z w_colors = dfr:w_colors
	endif
		
	for(i=0;i<numTraces;i+=1)
		traceStr = StringFromList(i,traceListStr)
		if (status & 2) // save current
			w_traces[i] = {traceStr}
			wave w_rgb = GetColor(graphStr, traceStr)
			w_colors[i][]={{w_rgb[0]},{w_rgb[1]},{w_rgb[2]}}
		else // check to see if a trace has been added since the list was created
			if (WaveExists(w_traces))
				FindValue /TEXT=traceStr/TXOP=4/Z w_traces
				if (V_Value == -1)
					w_traces[numpnts(w_traces)]={traceStr}
					wave w_rgb = GetColor(graphStr, traceStr)
					w_colors[DimSize(w_colors, 0)][] = {{w_rgb[0]},{w_rgb[1]},{w_rgb[2]}}
				endif
			endif
		endif
		if (status&1) // set to grey
			ModifyGraph /W=$graphStr/Z rgb($traceStr)=(56797,56797,56797)
		endif
	endfor
end

// use this with kRecolorTraces=1 to restore the saved color of a single trace
static function restoreTraceColor(graphStr, traceStr)
	string graphStr, traceStr
	
	DFREF dfr = getDFREF()
	wave /T w_traces = dfr:w_traces
	wave w_colors = dfr:w_colors
	
	FindValue /TEXT=traceStr/Z w_traces
	if (V_value >- 1)
		make /free w = {56797,56797,56797}
		w = abs(w - w_colors[V_Value][q])
		if (sum(w) < 1000)
			ModifyGraph /W=$graphStr/Z rgb($traceStr)=(39321,1,31457)
		else
			ModifyGraph /W=$graphStr/Z rgb($traceStr)=(w_colors[V_value][0],w_colors[V_value][1],w_colors[V_value][2])
		endif
	else
		ModifyGraph /W=$graphStr/Z rgb($traceStr)=(39321,1,31457)
	endif
end

static function /WAVE GetColor(graphStr, traceStr)
	string graphStr, traceStr
		
	variable c0, c1, c2
	string infoStr = TraceInfo(graphStr, traceStr, 0)
	variable startIndex = strsearch(infoStr, ";rgb(x)=", 0)
	if (startIndex == -1)
		return $"{}"
	endif
	sscanf infoStr[startIndex,strlen(infoStr)-1], ";rgb(x)=(%d,%d,%d*", c0, c1, c2
	Make /free w_rgb = {c0,c1,c2}
	return w_rgb
end

// execute BaselineSpline#makeSpectrum() to create a fake spectrum for demo
// creates a fake 'spectrum' containing Gaussian peaks
// superimposed on a noisy baseline
static function makeSpectrum([variable n])		
	n = ParamIsDefault(n) ? 1 : n	
	variable numPeaks = 5 // number of Gaussian peaks in the spectrum
	string strName
	variable a, b, c, d, i, j, height, position, width	
	Display /K=1
	for(i=0;i<n;i+=1)
		strName = UniqueName("demo", 1, 0)
		Make /N=1000 $(UniqueName("demo", 1, 0)) /WAVE=demo
		a = enoise(5); b = enoise(1e-3); c = enoise(1e-5); d = enoise(400)
		demo = gnoise(0.1) + a + b*x + c*(x-d)^2
		for(j=0;j<numPeaks;j+=1)
			height = 150 + enoise(50)
			position = 500 + enoise(400)
			width = 20 + enoise(10)
			demo += height*Gauss(x, position, width/sqrt(2))
		endfor
		AppendToGraph demo
	endfor
end

static function goodbye()
	
	if (IgorVersion() >= 8)			
		
		string msg = "Spline baselines are now included in the Baselines project.\r\r"

		if (strlen(functionInfo("baselines#DoSplineFit")))
			msg += "Please uninstall BaselineSpline and instead use the Baselines utility:\r\r"
			msg += "                   Analysis -> Baselines\r\rThanks!"
			doAlert 0, msg
			return 1
		endif

		if (strlen(functionInfo("updater#update")))
			msg += "Do you want to open IgorExchange project installer to look for the most recent version of Baselines?"
			DoAlert 1, msg
			if (v_flag == 1)
				execute "updater#makeInstallerPanel()"
				return 1
			endif
			return 0	
		endif
		
		msg += "Please uninstall BaselineSpline and install the most recent version (≥ 4.0) of the Baselines package.\r\r"
		msg += "Do you want to visit https://www.wavemetrics.com/project/Baselines?"
		
		DoAlert 1, msg
		if (v_flag == 1)
			BrowseURL /Z "https://www.wavemetrics.com/project/Baselines"
			return 1
		endif		
	endif
	
	return 0
end